/*
lsystem.c
using Lindenmayer formal languages for generating organic branching structures
Prim Andras 2002
*/

#include <ctype.h>
#include <stdlib.h>
#include <math.h>
#include "lsystem.h"
#include "myio.h"

  /* initialises the plant 'pl' according to lsystem 'ls' */
int LsInitPlant(plant *pl, lsystem *ls)
{
  debinfo("LsInitPlant()\n");
  int length;

  length = ls->axiomlength;
  pl->length = length;
  pl->ls = ls;
  pl->sentence = (char*) malloc(length + 1);
  memcpy(pl->sentence, ls->axiom, length + 1);

  pl->iteration = 0;
//  debinfo(ls->axiom);
  debinfo("LsInitPlant() end\n");
} // LsInitPlant()


  // LsDerive() : performs exactly one derivation on the plant according to
  // the rules of the plant's lsystem
int LsDerive(plant *pl)
{
    // actual values of token parameters
  float paramval[MAXPARAMETERNUM], odds;
  dynastring ds;
  char *sentenceptr, *lastptr, token;
  int i, ri, paramnum;
  lsystem *ls;

  debinfo("LsDerive()\n");

  ls = pl->ls;
  InitDS(&ds);
  sentenceptr = pl->sentence;

  debinfo(sentenceptr);
/*
for ( i = 0; i < pl->length; i++)
  printf("%x_%c_", sentenceptr[i], sentenceptr[i]);
printf("\n");
*/

  debinfo("LsDerive() 1\n");
  while(0 != *sentenceptr)
  {
//puts(sentenceptr);
//    lastptr = sentenceptr;
    token = *(sentenceptr++);
//printf("token <%c>", token);
    sentenceptr = LsGetParameters(sentenceptr, paramval, &paramnum);
      /* if a rule applies to the token and there is a 'p(x)'
       in the head, the rule is performed on the token only if odds < x.
       Otherwise LsHeadSat decreases odds with x. */
    odds = (float) rand() / RAND_MAX;
      /* applies the first rule for actual token, whose head is satisfied by the
         actual parameters  */
    for (ri = 0; ri < ls->rulenum; ri++)
    {
       if( (token == ls->rules[ri].token) &&
           LsHeadSat(ls, ri, paramval, &odds) )
       {
         LsApplyRuleDS(ls, ri, paramval, &ds);
	 ri = ls->rulenum + 1;
       }
    }

     // if no applicable rule found, simply copy the token with parameters
    if ( ri == ls->rulenum )
    {
//      i = sentenceptr - lastptr;
      AddCharToDS(&ds, token);
      if ( paramnum > 0)
      {
        AddCharToDS(&ds, '(');
        for (i = 0; i < paramnum-1; i++)
        {
          AddFloatToDS(&ds, paramval[i]);
          AddCharToDS(&ds, ',');
        }
        AddFloatToDS(&ds, paramval[i]);
        AddCharToDS(&ds, ')');
      }
    }

  } // while 0 != sentenceptr

  free(pl->sentence);
  pl->sentence = GetStringFromDS(&ds, &(pl->length));
  FreeDS(&ds);

  debinfo("LsDerive() end\n");
  return 1;
}  // LsDerive()


  /* gathers the parameter values into 'values' from 'sentenceptr'
     returns a pointer to the first character after the parameters
     sets 'paramnum' to the number of parameters read */
char *LsGetParameters(char *sentenceptr, float *values, int *paramnum)
{
  int i;
  float dec, val;

  debinfo("LsGetParameters()\n");

  memset(values, 0, sizeof(float) * MAXPARAMETERNUM);
  if ('(' != *sentenceptr)
  {
    *paramnum = 0;
    return sentenceptr;
  }

  i = 0;
    // find first number's beginning or parameterlist end
#ifdef FLOATTEXT
  while( !isdigit(*sentenceptr) && (')' != *sentenceptr))
#else
  while( !isdigit(*sentenceptr) && (FLOATESCAPE != *sentenceptr) && (')' !=
    *sentenceptr))
#endif
  {
    if (0 == *sentenceptr)
    {
      *paramnum = 0;
      return sentenceptr;
    }
    sentenceptr++;
  }

  while((*sentenceptr != ')') && (i < MAXPARAMETERNUM))
  {
#ifndef FLOATTEXT
    if(FLOATESCAPE == *sentenceptr)
    {
      sentenceptr++;
      memcpy(&val, sentenceptr, sizeof(float));
      sentenceptr += sizeof(float);
    }
    else
#endif
    {
      val = 0;
        // digits before decimal point
      while(isdigit(*sentenceptr))
      {
        val = val*10 + *sentenceptr - '0';
        sentenceptr++;
      }
      if('.' == *sentenceptr)
      {
        dec = 0.1;
          // skip decimal point
        sentenceptr++;
        while(isdigit(*sentenceptr))
        {
          val += dec * (*sentenceptr - '0');
          dec /= 10.0;
          sentenceptr++;
        }
      } // if there is decimal point in the number
    }

    values[i] = val;
    i++;

      // seek next number or parameters end
#ifdef FLOATTEXT
    while( !isdigit(*sentenceptr) && (')' != *sentenceptr))
#else
    while( !isdigit(*sentenceptr) && (FLOATESCAPE != *sentenceptr) && (')' !=
      *sentenceptr))
#endif
    {
      if (0 == *sentenceptr)
      {
        *paramnum = i;
        return sentenceptr;
      }
      sentenceptr++;
    }
  } // while(sentenceptr != ')')

    // set unspecified parameters to 0
  memset(&(values[i]), 0, sizeof(float) * (MAXPARAMETERNUM - i));

    // needed if the loop above exits due to i>= MAXPARAMETERNUM
  while(*sentenceptr != ')')
    sentenceptr++;
    // skip ')'
  sentenceptr++;

  debinfo("LsGetParameters() end\n");
  *paramnum = i;
  return sentenceptr;
} // LsGetParameters()


  /* returns true if 'ls's rule number rnum's head
     is satisfied by actual parameters in 'paramval' */
int LsHeadSat(lsystem *ls, int rulenum, float *paramval, float *odds)
{
/*debinfo2("LsHeadSet rulenum %d", rulenum);
debinfo2(" ls->rules[rulenum].prb %f ", ls->rules[rulenum].prb);
debinfo2(" odds %f\n", *odds); */
  if (*odds <= ls->rules[rulenum].prb)
    return 1;

  *odds -= ls->rules[rulenum].prb;
  return 0;
} // LsHeadSat()


  /* applies 'rlnum'th rule of 'ls' with actual parameters 'paramval', concats
     the result to 'ds'
     if 'rlnum' < 0, processes the axiom of ls */
int LsApplyRuleDS(lsystem *ls, int rlnum, float *paramval, dynastring *ds)
{
  rule *rl;
  char *body, *last, token;
  float value, dec, tmp, min, max, diff;
  int itmp, length, decpoint;
  enum {add, subtract, multiply, divide, assign} todo;

  debinfo("LsApplyRuleDS()\n");
//printf("%c paramval[0] %f\n", ls->rules[rlnum].token, paramval[0]);

  if(rlnum >= 0)
  {
    rl = &(ls->rules[rlnum]);
    body = rl->body;
  }
  else
  {
    rl = NULL;
    body = ls->axiom;
  }

  while (0 != *body)
  {
//    last = body;
    token = *(body++);

    AddCharToDS(ds, token);
    decpoint = 0;
    value = 0.0;
    tmp = 0.0;
    todo = assign;


      // if not reached end of string 'body' and token have parameter list
    if ((token != 0) && (*body == '('))
    {
      AddCharToDS(ds, '(');
        // evaluate parameters
      do
      {
        body++;
          /* at these characters we have to evaluate 'value'
             it is done without precedence yet
          */
        if(NULL != strchr("),+-*/", *body))
        {
//          if(tmp > 0.0)
//          {
            switch(todo)
            {
              case assign:
                value = tmp;
                break;
              case add:
                value += tmp;
                break;
              case subtract:
                value -= tmp;
                break;
              case multiply:
                value *= tmp;
                break;
              case divide:
                value /= tmp;
                break;
            } // switch(todo)
//          } // if (tmp > 0.0)
          tmp = 0.0;
        } // end if we have to evaluate

          // next parameter
        if((*body == ',') || (*body == ')'))
        {
          AddFloatToDS(ds, value);
          AddCharToDS(ds, *body);
          value = 0.0;
          tmp = 0.0;
          todo = assign;
          decpoint = 0;
        }
          // decimal point
        else if (*body == '.')
        {
          dec = 0.1;
          decpoint = 1;
        }
        else if(isdigit(*body))
        {
            // if we are after a decimal point
          if(decpoint)
          {
            tmp += dec * (*body - '0');
            dec /= 10.0;
          }
          else
            tmp = tmp * 10 + *body - '0';
        }
        else if (*body == '+')
          todo = add;
        else if (*body == '-')
          todo = subtract;
        else if (*body == '*')
          todo = multiply;
        else if (*body == '/')
          todo = divide;
          // referencing the derivable token's parameter
        else if (*body == '$')
        {
          itmp = 0;
          body++;
          while (isdigit(*body))
          {
            itmp = 10 * itmp + *body - '0';
            body++;
          }
          body--;


          tmp = (itmp > 0) && (itmp <= MAXPARAMETERNUM) ?
                 paramval[itmp-1] : 0.0;
        }
        else
        {
            // checking whether substring is a parameter of the lsystem
          for (itmp = 0; itmp < ls->paramnum; itmp ++)
          {
            if((length = strbegin(body, ls->params[itmp])) &&
             !isalnum(body[length]))
            {
              min = ls->paramvalmins[itmp];
              max = ls->paramvalmaxs[itmp];
              diff = max - min;
              tmp = (float) rand() / RAND_MAX;
              tmp = min + tmp * diff;
               // -1 since a body++ will be done anyway
              body += length - 1;
              break;
            }
          }
          /* at this point either *body was a parameter of the lsystem (in the
             if above), or a senseless character */
        }

      }while((')' != *body) && (0 != *body)); // parameter evaluation

      if (')' == *body)
        body++;

    }
  } //  while (0 != *body)
  debinfo("LsApplyRuleDS() end\n");

} // LsApplyRuleDS()


  /* Reads an lsystem from a file */
lsystem *LsReadFile(FILE *fil)
{
  lsystem *ls;
  rule *rl;
  char ch, token, *head, *body;
  float paramvalmin, paramvalmax, prb;
  float axiomparam[MAXPARAMETERNUM];
  int i, index, *pi, isnew;
  stack paramminstack, parammaxstack, paramstack, paramlengthstack;
  dynastring axiomds;

debinfo("LsReadFile 0\n");
  ls = (lsystem*) malloc(sizeof(lsystem));
  ls->leaf.trianglenum = 0;
  ls->leaf.triangles = NULL;
  ls->leaf.normals = NULL;
  ls->axiom = NULL;
  ls->rules = NULL;
  ls->rulenum = 0;
//  ls->specchars = NULL;
  ls->params = NULL;
  ls->paramstore = NULL;
  ls->paramvalmins = NULL;
  ls->paramvalmaxs = NULL;
  ls->paramnum = 0;

debinfo("LsReadFile 1\n");
  LsReadLeafdef(fil, &ls->leaf);

  fskipNLsorcomments(fil);

    //////////////// specchars
/*  if (NULL == (ls->specchars = freadstring(fil)))
  {
    errorinfo("reached EOF before the special character string");
    freeLS(ls);
    return NULL;
  }
  fskipline(fil);
  fskipNLsorcomments(fil);
debinfo("LsReadFile specchars  ");
debinfo(ls->specchars);
*/

    //////////////// parameters
  initstack(&paramstack);
  initstack(&paramlengthstack);
  initstack(&paramminstack);
  initstack(&parammaxstack);

  do
  {
    isnew = LsReadParam(fil, &paramstack, &paramlengthstack,
      &paramminstack, &parammaxstack);
    if (isnew)
    {
      ls->paramnum ++;
      fskipline(fil);
    }
  }while (isnew);
  ls->paramstore = getstack(&paramstack, &i);
  freestack(&paramstack);
  pi = (int*) getstackunsafe(&paramlengthstack, &i);
  ls->params = malloc(ls->paramnum*sizeof(char*));
  for (i = 0, index = 0; i < ls->paramnum; i++)
  {
    ls->params[i] = ls->paramstore + index;
      // +1 for the trailing zero
    index += pi[i] + 1;
  }
  freestack(&paramlengthstack);
  ls->paramvalmins = getstack(&paramminstack, &i);
  ls->paramvalmaxs = getstack(&parammaxstack, &i);
  freestack(&paramminstack);
  freestack(&parammaxstack);


debinfo("\nLsReadFile paramchs\n");

  fskipNLsorcomments(fil);

    //////////////// axiom
  if (NULL == (ls->axiom = freadstring(fil)))
  {
    errorinfo("reached EOF before the axiom string");
    freeLS(ls);
    return NULL;
  }
  fskipline(fil);
  fskipNLsorcomments(fil);

    // evaluates the expressions in axiom's parameters
//printf("before: <%s>\n", ls->axiom);
  InitDS(&axiomds);
  memset(axiomparam, 0, MAXPARAMETERNUM*sizeof(float));
  LsApplyRuleDS(ls, -1, axiomparam, &axiomds);
  free(ls->axiom);
  ls->axiom = GetStringFromDS(&axiomds,&ls->axiomlength);
  FreeDS(&axiomds);
debinfo("\nLsReadFile axiom ");
debinfo(ls->axiom);
//printf("after: <%s>\n", ls->axiom);


    ////////////////// rules
  do
  {
    LsReadRule(fil, &token, &head, &body, &prb);
    if (token)
    {
      i = ls->rulenum ++;
      ls->rules = realloc(ls->rules, ls->rulenum * sizeof(rule));

      rl = &(ls->rules[ls->rulenum - 1]);
      rl->token = token;
      rl->head = head;
      rl->body = body;
      rl->prb = prb;
debinfo2("prb %f\n", prb);
      fskipline(fil);
    }
  }while (token);

debinfo("\n LsReadFile end\n");
  return ls;

} // LsReadFile()

int LsReadLeafdef (FILE *fil, mesh *pmesh)
{
  triangle3f triangletmp, *ptriangle;
  stack trst;
  int linebegin, ch, thereistr = 1, trnum, i, j;
  char trsep[4] = {']', CR, LF, 0};

  initstack (&trst);
  fskipNLsorcomments(fil);
    /* this loop reads a line per round
     stops when reaches a blank line, or one, which is not a comment or
     a well formed triangle definition */
  do {
    linebegin = ftell(fil);
      // leading blanks
    do {
      ch = fgetc(fil);
    } while ((ch == ' ') || ('\t' == ch));

      // comment
    if (ch == '#')
      fskipline(fil);
      // not triangle definition
    else if (ch != '[')
    {
      thereistr = 0;
      if ((ch == LF) || (ch == CR))
        linebegin = ftell(fil);
    }
      // probably triangle definition
    else
    {
      for (i = 0; i < 3; i++)
        for (j = 0; j < 3; j++)
        {
            /* freadfloat returns nonzero if EOF, CR, LF or ']'
             reached before a number could be read */
          if (freadfloat(fil, &triangletmp[i][j], trsep))
          {
            thereistr = 0;
            i = 3; j = 3;
          }
        }
        // this means, that we successfully read a triangle
      if(thereistr)
      {
        push(&trst, triangletmp, sizeof(triangle3f));
        fskipline(fil);
      }
    }

  }while(thereistr);

  if(pmesh->trianglenum)
  {
    ptriangle = getstackunsafe(&trst, &trnum);
    pmesh->triangles =
      realloc(pmesh->triangles, pmesh->trianglenum*sizeof(triangle3f) + trnum);
    memcpy(&pmesh->triangles[pmesh->trianglenum], ptriangle, trnum);
    trnum /= sizeof(triangle3f);
    pmesh->normals =
      realloc(pmesh->normals, (pmesh->trianglenum  + trnum)*sizeof(vector3f));
    for (i = pmesh->trianglenum, j = pmesh->trianglenum + trnum; i < j; i++)
      GetNormal3f(pmesh->triangles[i], pmesh->normals[i]);
    pmesh->trianglenum += trnum;
  }
  else
  {
    pmesh->triangles = getstack(&trst, &trnum);
    trnum /= sizeof(triangle3f);
    pmesh->normals = malloc(trnum * sizeof(vector3f));
    for ( i = 0; i < trnum; i++)
    {
      GetNormal3f(pmesh->triangles[i], pmesh->normals[i]);
    }
    pmesh->trianglenum = trnum;
  }
  fseek(fil, linebegin, SEEK_SET);
  freestack (&trst);
  return 0;
}


  /* reads a parameter name/valuemin/valuemax triplet from 'fil', skipping
     comments
     pushes the name into paramstack, the (int) strlen(name) into lengthstack
     (float) valuemin and (float) valuemax into minstack and maxstack
     returns 1 if found 0 if not found */
int LsReadParam(FILE *fil, stack *paramstack, stack *lengthstack,
      stack *minstack, stack *maxstack)
{
  char ch2;
  int ch, i;
  float paramvalmin, paramvalmax, value, dec;

    // skip blanks
  do
  {
    ch = fgetc(fil);
  }while ( (' ' == ch) || ('\t' == ch) );
    // error
  if ((ch == EOF) || (ch == LF) || (ch == CR))
  {
    return 0;
  }

  i = 0;
  do
  {
    ch2 = (char)ch;
    push(paramstack, &ch2, 1);
    i++;
    ch = fgetc(fil);
  }while ( isalnum(ch) );
  ch2 = 0;
  push(paramstack, &ch2, 1);
  push(lengthstack, &i, sizeof(int));

    // error
  if ((ch == EOF) || (ch == LF) || (ch == CR))
  {
    return 0;
  }

    // we read two numbers
  for (i = 0; i < 2; i++)
  {
      /* skips almost anything before the number */
    do
    {
      ch = fgetc(fil);
      if ((EOF == ch) || (CR == ch) || (LF == ch))
      {
          // if we already read one number
        if (i)
        {
          push(minstack, &paramvalmin, sizeof(float));
          push(maxstack, &paramvalmin, sizeof(float));
          if (EOF != ch)
            fseek (fil, -1, SEEK_CUR);
          return 1;
        }
        else
        {
          return 0;
        }
      }
    }while (! isdigit(ch));

    value = 0.0;
    dec = 0.0;
      // reads the number
    do
    {
      if (ch == '.')
        dec = 0.1;
      else if (dec == 0.0)
        value = value*10 + ch - '0';
      else
      {
        value += dec * (ch - '0');
        dec /= 10;
      }
      ch = fgetc(fil);
    }while( isdigit(ch) || ('.' == ch));
    fseek (fil, -1, SEEK_CUR);

      // first number stored in paramvalmin
    if (!i)
      paramvalmin = value;
  } // for read two numbers

  if (value < paramvalmin)
  {
    paramvalmax = paramvalmin;
    paramvalmin = value;
  }
  else
  {
    paramvalmax = value;
  }

  push(minstack, &paramvalmin, sizeof(float));
  push(maxstack, &paramvalmax, sizeof(float));
  return 1;
} // LsReadParam()


  /* reads a rule from file 'fil'
     'token' is the token on which the rule applies, or 0 if an error rised
         while reading the rule
     'head' and 'body' will contain the head and body of the rule in newly
         allocated buffers */
void LsReadRule(FILE *fil, char *token, char **head, char **body, float *prb)
{
  long headbegin, headend, bodybegin, bodyend, length;
  enum {prehead, inhead, posthead, prebody, inbody, postbody} what = prehead;
  int ch, lastch, i, j;

  debinfo("LsReadRule 0\n");
    // skip blanks
  do
  {
    ch = fgetc(fil);
  }while ( (' ' == ch) || ('\t' == ch) );
    // error
  if ((ch == EOF) || (ch == LF) || (ch == CR))
  {
    *token = 0;
    return;
  }

  debinfo("LsReadRule 1\n");

    // le token
  *token = ch;

    // search for ':'
  do
  {
    ch = fgetc(fil);
    if (EOF == ch)
    {
      *token = 0;
      return;
    }
  }while ( ':' != ch );
  debinfo("LsReadRule 2\n");

  do
  {
    lastch = ch;
    ch = fgetc(fil);

    switch(ch)
    {
      case ' ':
      case '\t':
        break;
      case '>':
        if ( ((what == posthead) || (what == inhead))  &&  ('-' == lastch))
          what = prebody;
        break;
      case '"':
        if (what == inbody)
        {
          bodyend = ftell(fil) - 2;
          what = postbody;
        }
        else if (what == prebody)
        {
          bodybegin = ftell(fil);
          what = inbody;
        }
        break;
      case EOF:
        if (what == inbody)
        {
          bodyend = ftell(fil) - 2;
          what = postbody;
        }
        else
        {
          *token = 0;
          return;
        }
        break;
      case LF:
      case CR:
        break;
      default:
        if (what == prehead)
        {
          what = inhead;
          headbegin = ftell(fil) - 1;
          headend = headbegin;
        }
        else if ((what == inhead) && (ch != '-'))
        {
          headend = ftell(fil);
        }
    }
  }while (what != postbody);

  debinfo("LsReadRule 3\n");

  length = headend - headbegin + 1;
  length = length >= 0 ? length : 0;
  *head = (char*) malloc(length + 1);
  if (length)
  {
    fseek (fil, headbegin, SEEK_SET);
    fread(*head, 1, length, fil);
  }
  (*head)[length] = 0;

    /* exclude space , vertical tab, CR, LF from head and
      extracts probability value, if present */
  *prb = 2.0;
  for (i = 0, j = 0; i < length; i++)
  {
    ch = (*head)[i];
    if ('p' == ch)
    {
      while((i < length) && ('(' != (*head)[i])) i++;
      if (i < length)
        i += strnextfloat(&(*head)[i], prb);
      while((i < length) && (')' != (*head)[i])) i++;
    }
    else if ( !((ch == CR) || (ch == LF) || (ch == ' ') || (ch == '\t')) )
    {
      (*head)[j++] = ch;
    }
  }
  (*head)[j] = 0;
  debinfo("LsReadRule 3b\n");

  length = bodyend - bodybegin + 1;
  length = length >= 0 ? length : 0;

  *body = (char*) malloc(length + 1);
  if (length)
  {
    fseek (fil, bodybegin, SEEK_SET);
    fread(*body, 1, length, fil);
  }

    // exclude space , vertical tab, CR, LF from body
  for (i = 0, j = 0; i < length; i++)
  {
    ch = (*body)[i];
    if ( !((ch == CR) || (ch == LF) || (ch == ' ') || (ch == '\t')) )
    {
      (*body)[j++] = ch;
    }
  }
  (*body)[j] = 0;
  debinfo("LsReadRule 4\n");

} // LsReadRule();


  /* frees memory allocated by ls */
void freeLS(lsystem *ls)
{
  int i;
  rule *rl;
  if (NULL == ls)
    return;
  if (NULL != ls->axiom)
    free(ls->axiom);
  if (NULL != ls->rules)
  {
    for (i = 0; i < ls->rulenum; i++)
    {
      rl = &(ls->rules[i]);
      if (NULL != rl->head)
        free(rl->head);
      if (NULL != rl->body)
        free (rl->body);
    }
    free(ls->rules);
  }
  if(NULL != ls->params)
    free(ls->params);
  if(NULL != ls->paramstore)
    free(ls->paramstore);
  if (NULL != ls->paramvalmins)
    free(ls->paramvalmins);
  if (NULL != ls->paramvalmaxs)
    free(ls->paramvalmaxs);

} // freeLS()

 /* reads the next float number from str (skipping any character before the
first digit)
returns the (0 based) index of the first character _after_ the number read */
int strnextfloat(char *str, float *val)
{
debinfo("strnextfloat\n");
  int i = 0;
  float dec;
  *val = 0;
  while (!isdigit(str[i]))
  {
    if (0 == str[i])
    {
debinfo("strnextfloat eh\n");
      return i;
    }
    i++;
  }

  while (isdigit(str[i]))
    *val = *val * 10 + str[i++] - '0';

  if('.' == str[i])
  {
    dec = 0.1;
    i++;
    while(isdigit(str[i]))
    {
      *val += dec * (str[i++] - '0');
      dec /= 10;
    }
  }
debinfo("strnextfloat end\n");
  return i;
}

int LsPrintSentence(plant *pl)
{
  char *body;
  int i, j, indent=0, lin=0;

  body = pl->sentence;

  while (*body != 0)
  {
    if (*body == '[')
    {
      puts("");
      for (j = 0; j < indent; j++)
        putchar(' ');
      lin = indent;
      indent ++;
    }
    if (lin >= 70)
    {
      puts("");
      for (j = 0; j < indent; j++)
       putchar(' ');
      lin = indent;
    }

    putchar (*body);
    if (*body == ']')
    {
      puts("");
      indent--;
      for (j = 0; j < indent; j++)
        putchar(' ');
      lin = indent;
    }
    else if((*body == ')') || (*body == ','))
    {
      putchar(' ');
      lin++;
    }

    body ++;
    lin++;
  }
  puts("");
  return 0;
}

