/*
lsystem.h
using Lindenmayer formal languages for generating organic branching structures
only context free at the moment
Prim Andras 2002
*/
#ifndef LSYSTEM_H
#define LSYSTEM_H

#include <stdio.h>
#include "mystring.h"
#include "pastuff.h"
#include "mesh2.h"


  /* a token in an L system shall have maximum MAXPARAMETERNUM parameters
     this enables to store parameters in a static array */
#define MAXPARAMETERNUM  8

  // struct rule : one L system rule
typedef struct
{
    // probability of whether this rule is performed
  float prb;
    // the token on wich the rule replaces
  char token;
    // prerequisities for this rule to be used
  char *head;
    // the token will be replaced with this
  char *body;
    // length of the body string
//  int bodylength;
}rule;


typedef struct
{
    // triangles of the leaf
  mesh leaf;
    // the initial sentence
  char *axiom;
  int axiomlength;
    // array of rules
  rule *rules;
  int rulenum;
    // characters, which shall be replaced with themselves at derivation
//  char *specchars;
    // character representation and min, max values of parameters
  char **params;
  char *paramstore;
  float *paramvalmins, *paramvalmaxs;
    // number of parameters
  int paramnum;
}lsystem;

typedef struct
{
  lsystem *ls;
    // actual sentence
  char *sentence;
    // length of actual sentence
  int length;
    // actual iteration number
  int iteration;
}plant;


  /* initialises the plant 'pl' according to lsystem 'ls' */
int LsInitPlant(plant *pl, lsystem *ls);
  /* LsDerive() : performs exactly one derivation on the plant according to
     the rules of the plant's lsystem */
int LsDerive(plant *pl);
  /* gathers the parameter values into 'values' from 'sentenceptr'
     returns a pointer to the first character after the parameters
     sets 'paramnum' to the number of parameters read */
char *LsGetParameters(char *sentenceptr, float *values, int *paramnum);
  /* returns true if 'ls's rule number rnum's head
     is satisfied by actual parameters in 'paramval' */
int LsHeadSat(lsystem *ls, int rulenum, float *paramval, float *odds);
  /* applies 'rlnum'th rule of 'ls' with actual parameters 'paramval', concats
     the result to 'ds'
     if 'rlnum' < 0, processes the axiom of ls */
int LsApplyRuleDS(lsystem *ls, int rlnum, float *paramval, dynastring *ds);
  /* Reads an lsystem from a file, returns it in a newly allocated buffer
     format of the file should be something like
     <file> ::= <NLsorcomments><leafdef>
       <NLsorcomments><NL><specchars>
       <NLsorcomments><parameters>
       <NLsorcomments><NL><NLsorcomments><axiom>
       <NLsorcomments><NL><NLsorcomments><rules>
       <NLsorcomments><NL>

     <NL> ::= <blanks>\n
     <comment> ::= <blanks>'#'<comment text><NL>
     <NLorcomment> ::= <NL> | <comment>
     <NLsorcomments> ::= <NLorcomment><NLsorcomments> | NIL

     <leafdef> ::= <triangles><NL> | NIL
     <triangles> ::= <triangle><triangles> | <triangle>
     <triangle> ::= '['<vertex><vertex><vertex>']'<NLorcomment>
     <vertex> ::= <blanks><float><blanks>

     <axiom> ::= <blanks>'"'<characters>'"'<NL>

     <parameters> ::= <parameter><parameters> | NIL
     <parameter> ::= <blanks><string><nondigits><number><NLorcomment> |
             <blanks><character><nondigits><number>
             <nondigits><number><NLorcomment>

     <rules> ::= <rule><rules> | <rule>
     <rule> ::= <token><blanks>':'<blanks><head><blanks>'->'
       <body><NLorcomment>

     <token> ::= <character>
     <head> ::= '*' | 'p('<number>')'| NIL
     <body> ::= <blanks>'"'<characters>'"'

     special characters:
       space, vertical tab, CR, LF are excluded from <body> before processing
       in parameterlist @ffff can be a parameter, where @ is the FLOATESCAPE
         defined in mystring.h, and ffff is the byte representation of a float
         number
       'p' which is reserved for probability

      */
lsystem *LsReadFile(FILE *fil);

/* adds the triangles read from 'fil' to pmesh
 sets the file pointer after the <NL> that shows the end of the triangles
 or at the beginning of the first line which doesn't seem to be a triangle
 definition */
int LsReadLeafdef (FILE *fil, mesh *pmesh);

  /* reads a parameter name/valuemin/valuemax triplet from 'fil', skipping
     comments
     pushes the name into paramstack, the (int) strlen(name) into lengthstack
     (float) valuemin and (float) valuemax into minstack and maxstack
     returns 1 if found 0 if not found */
int LsReadParam(FILE *fil, stack *paramstack, stack *lengthstack,
      stack *minstack, stack *maxstack);

  /* reads a rule from file 'fil'
     'token' is the token on which the rule applies, or 0 if an error rised
         while reading the rule
     'head' and 'body' will contain the head and body of the rule in newly
         allocated buffers
     prb is the probability of the rule */
void LsReadRule(FILE *fil, char *token, char **head, char **body, float *prb);

  /* frees memory allocated by ls */
void freeLS(lsystem *ls);
 /* reads the next float number from str (skipping any character before the
first digit)
returns the (0 based) index of the first character _after_ the number read
if no float is found before the end of string, val=0 return value is the index
of the trailing zero */
int strnextfloat(char *str, float *val);

/* prints pl->sentence with indentation to stdout */
int LsPrintSentence(plant *pl);
#endif /* first inclusion of this header */
