/*
mesh2.h
Prim Andras 2002
*/

#ifndef MESH2_H
#define MESH2_H

#include "pagl.h"
#include "pastuff.h"
#include <stdio.h>

  // definitions of flag bits
  // common for mvertex, mtriangle, sectline
#define TOUCHED      1
  // sectline
#define SIDE         2
#define FIRSTSECT    4
  // vertex
#define SUBDIVVERTEX 2
  // triangle
#define NEIGHBOUR01  4
#define NEIGHBOUR02  8
#define NEIGHBOUR12 16

#define N01 4
#define N02 8
#define N12 16
  // end of definitions of flag bits

#define DRAW_SOLID 1
#define DRAW_WIRE  2
#define DRAW_DEBUG_WIRE 3

// default parameter of subdivision
#define DEFAULT_W      0.0
// weight of neighbour vertexes when counting VertexAvarage
#define AVARAGEWEIGHT  0.3

typedef struct _mesh {
  int trianglenum;
  triangle3f *triangles;
  vector3f *normals;
} mesh;


#define ENDTYPENUM 8
typedef enum {none, inner, A, B, C, AB, AC, BC} tendtype;
typedef struct _sectline {
    /* coordinates of endpoints of the section line */
  vertex3f endpoint[2];
    /* type of endpoints */
  tendtype endtype[2];
  vector3f normal;
    /* index of triangle to which this line belongs */
  int trindex;
  char flag;
} sectline;

typedef struct _mvertex {
  vertex3f coord;
  struct _mline *pfirstline;
    /* pfirstline->pvertex[vindex] points on this vertex */
  char vindex;
  char flag;
}mvertex;

typedef struct _mtriangle {
  mvertex *angles[3];
  vector3f normal;
    /* neighbours[0] = neighbour with angles[0] - angles[1] side
       neighbours[1] = neighbour with angles[0] - angles[2] side
       neighbours[2] = neighbour with angles[1] - angles[2] side
    */
  struct _mtriangle *neighbours[3];
  struct _mline *pline[3];
  int trindex;
  char flag;
} mtriangle;

typedef struct _mline {
  mvertex *pvertex[2];
    /* pnextline[i] is the next line of pvertex[i] */
  struct _mline *pnextline[2];
    /* pnextline[i]->pvertex[vindex] points to pvertex[i] */
  char vindex[2];
  mtriangle *ptriangle[2];
  char trside[2];
  char flag;
}mline;

typedef struct _vertexindexnode {
  mvertex *item;
    /* the vertex (x,y,z) is in the offspring with the index
      (x > item.x ? 4 : 0) + (y > item.y ? 2 : 0) + (z > item.z ? 1 : 0)
    */
  struct _vertexindexnode *offsprings[8];
} vertexindexnode;


typedef struct _mesh2 {
  store *trianglestore;
  store *vertexstore;
  store *linestore;
  vertexindexnode *vertexindexroot;
  matrix4x4f transformmatrix;
  vector3f color;
} mesh2;



int InitMesh(mesh2 *pmesh);
 /* It inits the new mesh as well */
mesh2* NewMesh();
  /* does not delete the mesh2 structure */
int FreeStaticMesh(mesh2 *pmesh);
int FreeMesh(mesh2 *pmesh);
  /* creates new mesh and copies the _pointers_ of pmesh in it */
mesh2 *CopyMeshUnsafe(mesh2 *pmesh);
  /* frees node and it's offsprings recursively */
int FreeVertexIndexNode(vertexindexnode *pnode);
  /* exchange the contets of the two meshes */
int ExchangeMesh(mesh2 *pamesh, mesh2 *pbmesh);

int MeshToOldFormat(mesh2 *pmeshin, mesh *pmeshout);
  /* puts the triangle with index 'trindex' of pmesh into 'ptriangle' in
  triangle3f form */
int GetMeshTriangle3f (mesh2 *pmesh, int trindex, triangle3f ptriangle);
  /* puts the normal with index 'nindex' of pmesh into 'pnormal' in
  normal3f form */
int GetMeshNormal3f (mesh2 *pmesh, int nindex, vector3f pnormal);
int GetMeshTriangleAndNormal3f (mesh2 *pmesh, int nindex, triangle3f ptriangle,
  vector3f pnormal);
int GetMeshLine3f(mesh2 *pmesh, int index, line3f pline);
  /* returns the number of triangles in pmesh
  puts the vertexes and normals of the triangles in pmesh into 'vertexes'
  and 'normals' three for each triangle (so that it can easily handled
  by OpenGL) */
int GetMeshTrianglesAndNormals3f(mesh2 *pmesh, vertex3f **vertexes, vector3f
**normals);

#define GetMeshVertexNum(a) getstoreitemnum(*(a)->vertexstore)
#define GetMeshTriangleNum(a) getstoreitemnum(*(a)->trianglestore)
#define GetMeshLineNum(a) getstoreitemnum(*(a)->linestore)
  /* adds the triangle to the mesh (allocates space for it, and fills
  values) */
int AddTriangleToMesh(mesh2 *pmesh, triangle3f ptriangle);
  /* inserts a new mline into pmesh between ptriangle->angles[a] and
  ptriangle->angles[b] */
int NewMeshLine(mesh2 *pmesh, mtriangle *ptriangle, int a, int b);
  /* adds a polygon to 'pmesh' defined by it's sides in 'pline'
  number of sides = 'linenum'
  so far 'pline[i]->endpoint[1]' is expected to be
 'pline[i+1]->endpoint[0]' */
int AddPolygonToMesh(mesh2 *pmesh, sectline *pline, int
linenum, vector3f normal);
  /* if mesh already contains a vertex similar to pvertex, returns it's address
   else inserts pvertex into the mesh, and returns the pointer
   it uses the space of pvertex in neither case  */
mvertex *VertexPointer(mesh2 *pmesh, vertex3f pvertex);
  /* finds the triangles neighbour on the angles[i] angles[j] side,
  sets the pointer triangle.neighbours on it, and sets the neighbours
  pointer on itself as well */
int FindNeighbourTriangleAndAdd (mesh2 *pmesh, mtriangle *ptriangle,
      int i, int j);
   /* applies 'mask' on every triangle's flag in 'pmesh' with bitwise and */
int AllTriangleFlagAndMask(mesh2 *pmesh, char mask);
int SetTriangleFlag(mesh2 *pmesh, int trindex, char mask);
int ResetTriangleFlag(mesh2 *pmesh, int trindex, char mask);
/* puts the subdivided 'pmeshin' into 'pmeshout'
lines shorter than 'minlength' won't be divided
'w' is the parameter of subdivision weights */
int SubDivide(mesh2 *pmeshin, mesh2 *pmeshout, float minlength, float w);
/* counts the sub division vertex of pline, and puts into 'vertex' */
int CountDivisionVertex(mesh2 *pmesh, mline *pline, vertex3f vertex, float w);
/* adds subdivision triangles to pmesh
 subdivided triangle is 'triangle'
 pvflag[0],[1],[2] & TOUCHED = true if there is a division vertex on
  'triangle's side 01, 02, 12 respectively
 new vertexes on triangles side 01,02,12 are in pdivvertex[0],[1],[2]
 */
int AddSubDivTriangles(mesh2 *pmesh, triangle3f triangle, vertex3f *pdivvertex,
      char *pvflag);
/* moves pmvertex->coord to be the avarage of it's neighbour vertexes */
int VertexAvarage(mvertex *pmvertex, float weight);

/* writes the triangles of 'pmesh' to a file in Renderman Interface
 Bytestream format
 puts 'linebegin' at the beginning of each written line */
int MeshToRib(mesh2 *pmesh, FILE* fil, const char* linebegin);

#ifdef WITHGL
int DrawMeshGL(mesh2 *pmesh, int type);
int DrawLineAndNormalGL (sectline *pline);
#endif // WITHGL

#endif // first inclusion of this header
